package com.coob.sjts.common.stUtil;

import org.bouncycastle.asn1.gm.GMObjectIdentifiers;
import org.bouncycastle.crypto.engines.SM2Engine;
import org.bouncycastle.crypto.params.ECDomainParameters;
import org.bouncycastle.crypto.params.ECPrivateKeyParameters;
import org.bouncycastle.crypto.params.ECPublicKeyParameters;
import org.bouncycastle.crypto.params.ParametersWithRandom;
import org.bouncycastle.jcajce.provider.asymmetric.ec.BCECPrivateKey;
import org.bouncycastle.jcajce.provider.asymmetric.ec.BCECPublicKey;
import org.bouncycastle.jce.provider.BouncyCastleProvider;

import java.nio.charset.StandardCharsets;
import java.security.*;
import java.security.spec.ECGenParameterSpec;
import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.X509EncodedKeySpec;
import java.util.Base64;


/**
 * SM2工具类
 * <p>
 * Created On 2022/3/7
 *
 * @author 苦行僧
 */
public final class Sm2Util {
    private static final String EC = "EC";

    private Sm2Util() {
    }

    /**
     * 签名
     *
     * @param input             待签名数据
     * @param encodedPrivateKey 私钥字符串
     * @return 签名字符串
     */
    public static String sign(String input, String encodedPrivateKey) throws Exception {
        Signature signature = getInstance();
        signature.initSign(privateKey(encodedPrivateKey));
        signature.update(input.getBytes(StandardCharsets.UTF_8));
        return new String(Base64.getEncoder().encode(signature.sign()), StandardCharsets.UTF_8);
    }

    /**
     * 验签
     *
     * @param input            待签名数据
     * @param encodedPublicKey 公钥字符串
     * @param sign             签名
     * @return 是否正确
     */
    public static boolean verify(String input, String encodedPublicKey, String sign) throws Exception {
        final Signature signature = getInstance();
        signature.initVerify(publicKey(encodedPublicKey));
        signature.update(input.getBytes(StandardCharsets.UTF_8));
        return signature.verify(Base64.getDecoder().decode(sign));
    }

    /**
     * 加密
     *
     * @param input            待加密数据
     * @param encodedPublicKey 公钥字符串
     * @return 密文
     */
    public static String encrypt(String input, String encodedPublicKey) throws Exception {
        BCECPublicKey publicKey = (BCECPublicKey) publicKey(encodedPublicKey);
        SM2Engine sm2Engine = new SM2Engine();
        sm2Engine.init(true,
                new ParametersWithRandom(
                        new ECPublicKeyParameters(
                                publicKey.getQ(),
                                new ECDomainParameters(
                                        publicKey.getParameters().getCurve(),
                                        publicKey.getParameters().getG(),
                                        publicKey.getParameters().getN()
                                )
                        ),
                        new SecureRandom()
                )
        );

        byte[] inputBytes = input.getBytes(StandardCharsets.UTF_8);
        byte[] cipherBytes = sm2Engine.processBlock(inputBytes, 0, inputBytes.length);
        return new String(Base64.getEncoder().encode(cipherBytes), StandardCharsets.UTF_8);
    }

    /**
     * 解密
     *
     * @param input             待解密数据
     * @param encodedPrivateKey 私钥字符串
     * @return 明文
     */
    public static String decrypt(String input, String encodedPrivateKey) throws Exception {
        BCECPrivateKey privateKey = (BCECPrivateKey) privateKey(encodedPrivateKey);
        SM2Engine sm2Engine = new SM2Engine();
        sm2Engine.init(false,
                new ECPrivateKeyParameters(
                        privateKey.getD(),
                        new ECDomainParameters(
                                privateKey.getParameters().getCurve(),
                                privateKey.getParameters().getG(),
                                privateKey.getParameters().getN()
                        )
                )
        );

        byte[] inputBytes = Base64.getDecoder().decode(input);
        byte[] plainBytes = sm2Engine.processBlock(inputBytes, 0, inputBytes.length);
        return new String(plainBytes, StandardCharsets.UTF_8);
    }

    /**
     * 获取实例
     */
    public static Signature getInstance() throws NoSuchAlgorithmException {
        return Signature.getInstance(GMObjectIdentifiers.sm2sign_with_sm3.toString(), new BouncyCastleProvider());
    }

    /**
     * 获取公钥
     */
    private static PublicKey publicKey(String encodedPublicKey) throws Exception {
        return KeyFactory.getInstance(EC, new BouncyCastleProvider())
                .generatePublic(new X509EncodedKeySpec(Base64.getDecoder().decode(encodedPublicKey)));
    }

    /**
     * 获取私钥
     */
    private static PrivateKey privateKey(String encodedPrivateKey) throws Exception {
        return KeyFactory.getInstance(EC, new BouncyCastleProvider())
                .generatePrivate(new PKCS8EncodedKeySpec(Base64.getDecoder().decode(encodedPrivateKey)));
    }

    /**
     * 生成公私钥对
     */
    public static KeyPair generateKeyPair() throws Exception {
        // 获取SM2椭圆曲线的参数
        final ECGenParameterSpec sm2Spec = new ECGenParameterSpec("sm2p256v1");
        final KeyPairGenerator kpg = KeyPairGenerator.getInstance("EC", new BouncyCastleProvider());
        kpg.initialize(sm2Spec);
        kpg.initialize(sm2Spec, new SecureRandom());
        return kpg.generateKeyPair();
    }

}
